from pathlib import Path


def delete_folder(p: Path) -> None:
    for sub in p.iterdir():
        if sub.is_dir():
            delete_folder(sub)
        else:
            sub.unlink()
    p.rmdir()



from functools import wraps
from typing import Any, Callable, TypeVar

WrappedFn = TypeVar("WrappedFn", bound=Callable[..., Any])


class Asset:
    def __init__(self, fn: WrappedFn) -> None:
        self.fn = fn

    def __call__(self) -> Any:
        return self.fn()

    def __name__(self) -> str:
        return self.fn.__name__


def asset(fn: WrappedFn) -> Asset:
    @wraps(fn)
    def asset_wrapper() -> Asset:
        return Asset(fn)

    return asset_wrapper()



import warnings
from pathlib import Path

import duckdb
import pandas as pd
from fastavro import reader


from daglib.utils.path import delete_folder


class MetaDB:
    def __init__(self) -> None:
        self._meta_dir = Path("meta/")
        if not self._meta_dir.exists():
            self._meta_dir.mkdir()

        self._profiling_dir = self._meta_dir / Path("profiling")
        if not self._profiling_dir.exists():
            self._profiling_dir.mkdir()

        profiling_record_paths = list(self._profiling_dir.rglob("*.avro"))

        profiling_records = []
        for p in profiling_record_paths:
            with open(p, "rb") as fp:
                for record in reader(fp):
                    profiling_records.append(record)

        self.profiling = pd.DataFrame(profiling_records)

    def query(self, query: str) -> pd.DataFrame | None:
        profiling = self.profiling  # noqa: F841
        con = duckdb.connect()
        try:
            df = con.execute(query).df()
            return df
        except RuntimeError:
            warnings.warn("No profiling records found")
            return pd.DataFrame()
        finally:
            con.close()

    def drop(self) -> None:
        try:
            delete_folder(self._meta_dir)
        except FileNotFoundError:
            pass
        self.profiling = self.profiling.head(0)



import time
from pathlib import Path
from typing import Any, TypeVar, Callable

from fastavro import writer, parse_schema

WrappedFn = TypeVar("WrappedFn", bound=Callable[..., Any])

_PROFILING_SCHEMA = parse_schema(
    {
        "doc": "daglib profiling stats",
        "name": "tprofile",
        "namespace": "daglib.profiling",
        "type": "record",
        "fields": [
            {"name": "dag_name", "type": "string"},
            {"name": "dag_description", "type": "string"},
            {"name": "run_id", "type": "string"},
            {"name": "task_name", "type": "string"},
            {"name": "task_description", "type": "string"},
            {"name": "task_runtime", "type": "float"},
        ],
    }
)


def profiling_wrapper(
    dag_name: str,
    dag_description: str,
    run_id: str,
    task_name: str,
    task_description: str,
    profile_dir: str,
    fn: WrappedFn,
    *args: Any,
    **kwargs: Any,
) -> Any:
    # noinspection PyShadowingNames
    def profiled_task(*args: Any, **kwargs: Any) -> Any:
        start = time.time()
        result = fn(*args, **kwargs)
        end = time.time()
        run_time = end - start
        record = {
            "dag_name": dag_name,
            "dag_description": dag_description,
            "run_id": run_id,
            "task_name": task_name,
            "task_description": task_description,
            "task_runtime": run_time,
        }
        out_dir = Path(profile_dir) / Path(dag_name)
        if not out_dir.parent.parent.exists():  # create meta/ if not exists
            out_dir.parent.parent.mkdir()
        if not out_dir.parent.exists():  # create meta/profiling if not exists
            out_dir.parent.mkdir()
        if not out_dir.exists():
            out_dir.mkdir()
        out_path = out_dir / Path(f"{run_id}.avro")
        if not out_path.exists():
            with open(out_path, "wb") as fp:
                writer(fp, _PROFILING_SCHEMA, records=[record])
        else:
            with open(out_path, "a+b") as fp:
                writer(fp, _PROFILING_SCHEMA, records=[record])
        return result

    return profiled_task



from typing import Any, Iterable, TypeVar, Callable

from daglib.core.profiling import profiling_wrapper

WrappedFn = TypeVar("WrappedFn", bound=Callable[..., Any])


class Task:
    def __init__(
        self,
        dag_name: str,
        dag_description: str,
        run_id: str,
        profile: bool,
        profile_dir: str,
        fn: WrappedFn,
        inputs: Any | Iterable[Any] = (),
        suffix: str | None = None,
        name_override: str | None = None,
    ) -> None:
        name = name_override if name_override else fn.__name__
        self.name = name if not suffix else f"{name} {suffix}"
        self.description = fn.__doc__ or ""

        if inputs == () and fn.__code__.co_argcount > 0:
            self.inputs = fn.__code__.co_varnames[: fn.__code__.co_argcount]
        elif not isinstance(inputs, Iterable) or isinstance(inputs, str):
            self.inputs = tuple([inputs])
        else:
            self.inputs = tuple(inputs)
        if not profile:
            self.fn = fn
        if profile:
            self.fn = profiling_wrapper(
                dag_name, dag_description, run_id, self.name, self.description, profile_dir, fn, *inputs
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['daglib', 'daglib.core', 'daglib.utils']

package_data = \
{'': ['*']}

install_requires = \
['dask>=2022.6.0,<2023.0.0',
 'duckdb>=0.4.0,<0.5.0',
 'fastavro>=1.5.3,<2.0.0',
 'numpy>=1.23.0,<2.0.0',
 'tenacity>=8.0.1,<9.0.0']

extras_require = \
{'graphviz': ['graphviz>=0.20,<0.21']}

setup_kwargs = {
    'name': 'daglib',
    'version': '0.3.2',
    'description': 'Lightweight DAG composition framework',
    'long_description': '# ⚗️ Daglib - Lightweight DAG composition framework\n\n[![PyPI version](https://badge.fury.io/py/daglib.svg)](https://badge.fury.io/py/daglib)\n[![PyPI - Downloads](https://img.shields.io/pypi/dm/daglib)](https://pypi.org/project/daglib/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/daglib.svg)](https://pypi.org/project/daglib/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n[![Checked with mypy](https://img.shields.io/badge/mypy-checked-blue.svg)](https://mypy.readthedocs.io/en/stable/)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n\nDaglib is a lightweight alternative to Airflow and other orchestration engines. It is meant to run on a single machine and comes with many great features out of the box like task I/O, dynamic task generation, and simple testing and deployment.\n\nIt can run as a standalone application or be embedded in another application to enable more complex use cases like event-driven workflows, conditional workflows, and more.\n\nSee documentation at https://mharrisb1.github.io/daglib/\n\n# Installation\n\n```shell\npip install daglib\n```\n\n# Create your first DAG\n\n\n```python\nimport daglib\n\ndag = daglib.Dag()\n\n\n@dag.task()\ndef task_1a():\n    return "Hello"\n\n\n@dag.task()\ndef task_1b():\n    return "world!"\n\n\n@dag.task(final=True)\ndef tassk_2(task_1a, task_1b):\n    print(f"{task_1a}, {task_1b}")\n\n\ndag.run()\n```\n\n    Hello, world!\n\n\n# Beyond the "Hello, world!" example\n\nFor a more involved example, we will create a small ETL pipeline that takes data from four source tables and creates a single reporting table. The data is driver-level information from the current 2022 Formula 1 season. The output will be a pivot table for team-level metrics.\n\n## Source Tables\n\n1. Team - Which team the driver belongs to for the season\n2. Points - Current total Driver\'s World Championship points for each driver for the season\n3. Wins - Current number of wins for each driver for the season\n4. Podiums - Current number of times the driver finished in the top 3 for the season\n\n\n```python\nimport pandas as pd\nimport daglib\n\n# Ignore. Used to render the DataFrame correctly in the README\npd.set_option("display.notebook_repr_html", False)\n\ndag = daglib.Dag()\n\n\n@dag.task()\ndef team():\n    return pd.DataFrame(dict(\n        driver=["Max", "Charles", "Lewis", "Sergio", "Carlos", "George"],\n        team=["Red Bull", "Ferrari", "Mercedes", "Red Bull", "Ferrari", "Mercedes"],\n    )).set_index("driver")\n\n\n@dag.task()\ndef points():\n    return pd.DataFrame(dict(\n        driver=["Max", "Charles", "Lewis", "Sergio", "Carlos", "George"],\n        points=[175, 126, 77, 129, 102, 111]\n    )).set_index("driver")\n\n\n@dag.task()\ndef wins():\n    return pd.DataFrame(dict(\n        driver=["Max", "Charles", "Lewis", "Sergio", "Carlos", "George"],\n        wins=[6, 2, 0, 1, 0, 0]\n    )).set_index("driver")\n\n\n@dag.task()\ndef podiums():\n    return pd.DataFrame(dict(\n        driver=["Max", "Charles", "Lewis", "Sergio", "Carlos", "George"],\n        podiums=[7, 4, 2, 5, 5, 3]\n    )).set_index("driver")\n\n\n@dag.task()\ndef driver_metrics(team, points, wins, podiums):\n    return team.join(points).join(wins).join(podiums)\n\n\n@dag.task(final=True)\ndef team_metrics(driver_metrics):\n    return driver_metrics.groupby("team").sum().sort_values("points", ascending=False)\n\n\ndag.run()\n```\n\n\n\n\n              points  wins  podiums\n    team\n    Red Bull     304     7       12\n    Ferrari      228     2        9\n    Mercedes     188     0        5\n\n\n\n## Task Graph Visualization\n\nThe DAG we created above will create a task graph that looks like the following\n\n![task graph](https://storage.googleapis.com/daglib-image-assets/example-dag.png)\n',
    'author': 'Michael Harris',
    'author_email': 'mharris@luabase.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mharrisb1/daglib',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import uuid
from typing import Any, Callable, Iterable, TypeVar

import numpy as np
from dask.delayed import Delayed
from dask.optimization import cull

from daglib.core.task import Task
from daglib.core.assets import Asset
from daglib.exceptions import TaskBuildError

WrappedFn = TypeVar("WrappedFn", bound=Callable[..., Any])


def chunk(arr: Iterable[Any], n: int) -> list[list[Any]]:
    return [list(sub) for sub in np.array_split(list(arr), n)]


def get(chunked_arr: list[list[Any]], i: int) -> list[Any]:
    try:
        return chunked_arr[i]
    except IndexError:
        return []


def find_keys(search_str: str, layers: dict[str, tuple[Any, ...]]) -> list[str] | str:
    keys = list(layers.keys())
    keys = list(filter(lambda k: k.split(" ")[0] == search_str, keys))
    if len(keys) > 1:
        return list(filter(lambda k: len(k.split(" ")) > 1 and "chunk" not in k.split(" "), keys))
    return keys[0]


def format_dag_name(s: str) -> str:
    return "".join(x for x in s if x.isalpha()).lower()


class Dag:
    def __init__(
        self,
        name: str | None = None,
        description: str = "",
        profile: bool = False,
    ) -> None:
        if not name:
            name = format_dag_name(str(uuid.uuid4()))
        self.name = format_dag_name(name)
        self.description = description
        self.profile = profile
        self.profile_dir = "meta/profiling/"
        self.run_id = "run_" + str(uuid.uuid1()).replace("-", "")[:9]
        self._tasks: list[Task] = []
        self._keys: list[str] = []

    def _register_task(
        self,
        fn: WrappedFn,
        inputs: Any | Iterable[Any] = (),
        suffix: str | None = None,
        final: bool = False,
        name_override: str | None = None,
        depends_on: list[WrappedFn | str] | None = None,
    ) -> Task:
        if depends_on:
            inputs = [*inputs, *[t.__name__ if callable(t) else t for t in depends_on]]
        task = Task(
            self.name, self.description, self.run_id, self.profile, self.profile_dir, fn, inputs, suffix, name_override
        )
        self._tasks.append(task)
        if final:
            self._keys.append(task.name)
        return task

    def _register_map_to_task(
        self,
        fn: WrappedFn,
        final: bool = False,
        joins: str | WrappedFn | None = None,
        map_to: list[Any] | str | WrappedFn | None = None,
        result_chunks: int | None = None,
        depends_on: list[WrappedFn | str] | None = None,
    ) -> None:
        if not map_to:
            map_to = []
        if joins:
            raise TaskBuildError("Task cannot have both map_to and joins specified. Choose one")
        if result_chunks:
            raise TaskBuildError("Task cannot have both map_to and result_chunks specified. Choose one")
        if callable(map_to):
            map_to = map_to.__name__
        if isinstance(map_to, str):
            map_to = [t.name for t in self._tasks if t.name.split(" ")[0] == map_to and len(t.name.split(" ")) > 1]
        for i, v in enumerate(map_to):
            self._register_task(fn, v, str(i), final, None, depends_on)

    def _register_joining_task(
        self,
        fn: WrappedFn,
        final: bool = False,
        joins: str | WrappedFn | None = None,
        map_to: list[Any] | str | WrappedFn | None = None,
        depends_on: list[WrappedFn | str] | None = None,
    ) -> Task:
        if map_to:
            raise TaskBuildError("Task cannot have both joins and map_to specified. Choose one")
        if callable(joins):
            joins = joins.__name__
        joined_tasks = [t.name for t in self._tasks if t.name.split(" ")[0] == joins and len(t.name.split(" ")) > 1]
        return self._register_task(fn, joined_tasks, None, final, None, depends_on)

    def _register_chunked_task(
        self,
        fn: WrappedFn,
        final: bool = False,
        joins: str | WrappedFn | None = None,
        map_to: list[Any] | str | WrappedFn | None = None,
        result_chunks: int | None = None,
        depends_on: list[WrappedFn | str] | None = None,
    ) -> None:
        if result_chunks is not None:
            if map_to:
                raise TaskBuildError("Task cannot have both result_chunks and map_to specified. Choose one")
            if joins:
                task = self._register_joining_task(fn, False, joins, None)
            else:
                task = self._register_task(fn, depends_on=depends_on)
            chunk_task = self._register_task(
                chunk, (task.name, result_chunks), task.name, depends_on=depends_on
            )  # type: ignore
            for n in range(result_chunks):
                self._register_task(get, (chunk_task.name, n), str(n), final, task.name, depends_on)  # type: ignore

    def task(
        self,
        final: bool = False,
        joins: str | WrappedFn | None = None,
        map_to: list[Any] | str | WrappedFn | None = None,
        result_chunks: int | None = None,
        depends_on: list[WrappedFn | str] | None = None,
    ) -> Callable[[WrappedFn], Callable[[WrappedFn], WrappedFn]]:
        def register(fn: WrappedFn) -> WrappedFn:
            if joins and not result_chunks:
                self._register_joining_task(fn, final, joins, map_to, depends_on)
            if map_to:
                self._register_map_to_task(fn, final, joins, map_to, result_chunks, depends_on)
            if result_chunks:
                self._register_chunked_task(fn, final, joins, map_to, result_chunks, depends_on)
            if not any([joins, map_to, result_chunks]):
                self._register_task(fn, (), None, final, None, depends_on)
            return fn

        return register

    @property
    def layers(self) -> dict[str, tuple[Any, ...]]:  # pragma: no cover
        layers = {task.name: tuple([task.fn, *task.inputs]) for task in self._tasks}
        return layers

    @property
    def final_tasks(self) -> list[str]:
        return self._keys

    def materialize(self, to_step: str | Callable[..., Any] | None = None, optimize: bool = False) -> Delayed:
        keys = self._keys[0] if len(self._keys) == 1 else self._keys
        layers = self.layers
        if to_step:
            if callable(to_step):
                to_step = to_step.__name__
            keys = find_keys(to_step, layers)
            optimize = True
        if optimize:
            layers, _ = cull(layers, keys)
        return Delayed(keys, layers)

    def add_subdag(self, other: Dag, keep_finals: bool = False) -> None:
        self._tasks += other._tasks
        if keep_finals:
            self._keys += other._keys

    def add_asset(self, asset: Asset) -> None:
        task = Task("", "", "", False, "", asset.fn)
        self._tasks.append(task)

    def run(self, to_step: str | Callable[..., Any] | None = None, optimize: bool = False) -> Any:
        return self.materialize(to_step, optimize).compute()

    # noinspection PyShadowingBuiltins
    def visualize(
        self,
        to_step: str | Callable[..., Any] | None = None,
        optimize: bool = False,
        filename: str | None = None,
        format: str | None = None,
        **kwargs: Any,
    ) -> Any:
        return self.materialize(to_step, optimize).visualize(filename, format, **kwargs)
